package io.tswf.minecraft.metamorphosis.cpw.modlauncher

import com.google.auto.service.AutoService
import cpw.mods.jarhandling.SecureJar
import cpw.mods.modlauncher.api.*
import groovy.transform.CompileStatic
import groovy.transform.Memoized
import groovy.util.logging.Slf4j
import io.tswf.minecraft.metamorphosis.core.remap.TinyRemappingAdapter
import io.tswf.minecraft.metamorphosis.core.transforming.MetamorphosisTransforms
import org.objectweb.asm.tree.ClassNode

@AutoService(ITransformationService)
@Slf4j
@CompileStatic
class MetamorphosisTransformationService implements ITransformationService {

    @Override
    String name() {
        return this.class.name
    }

    @Override
    void initialize(IEnvironment iEnvironment) {
        log.info("${this.class.simpleName}.initialize")
    }

    @Override
    void onLoad(IEnvironment iEnvironment, Set<String> set) throws IncompatibleEnvironmentException {
        log.info("${this.class.simpleName}.onLoad")
    }

    @Memoized
    @Override
    List<ITransformer> transformers() {
        _ensureLoaded()
        return Collections.<ITransformer>singletonList(new MetamorphosisTransformer())
    }

    @Memoized
    String _ensureLoaded() {
        delegateToObjectClassloader(this) {
            def result = MetamorphosisTransforms.typesToTransform
            log.info("Initialized with ${result.size()} transformation targets and remapping toggled to ${TinyRemappingAdapter.isRemappingEnabled}")
        }
    }

    static void delegateToObjectClassloader(Object obj, Closure action) {
        def oldClassloader = ClassLoader.contextClassloader()
        try {
            Thread.currentThread().contextClassLoader = obj.class.classLoader
            log.info("Temporary switched context classloader to ${Thread.currentThread().contextClassLoader}")
            action()
        } finally {
            Thread.currentThread().contextClassLoader = oldClassloader
        }
    }

    @Override
    List<Resource> completeScan(IModuleLayerManager layerManager) {
        def result = new ArrayList<Resource>()

        delegateToObjectClassloader(this) {
            def modsDir = new File("./mods-blockventure")
            if (modsDir.exists()) {
                def modsList = modsDir.walkSubFiles()
                        .filter { it.nameExtension == "jar" }
                        .distinctBy { it.absoluteCanonicalFile.path }
                        .toList()

                def cpwSecureJars = modsList
                        .map { SecureJar.from(it.toPath()) }

                log.info("Found custom mods folder: '${modsDir.absoluteCanonicalFile.path}' and ${modsList.size()} mod files in!")
                result = [ new Resource(IModuleLayerManager.Layer.GAME, cpwSecureJars) ]
            }
        }

        return result
    }

    @CompileStatic
    static class MetamorphosisTransformer implements ITransformer<ClassNode> {

        @Override
        ClassNode transform(ClassNode classNode, ITransformerVotingContext iTransformerVotingContext) {
            delegateToObjectClassloader(this) {
                MetamorphosisTransforms.transformClassNode(classNode)
            }
            return classNode
        }

        @Override
        TransformerVoteResult castVote(ITransformerVotingContext context) {
            return TransformerVoteResult.YES
        }

        @Memoized
        @Override
        Set<Target> targets() {
            return MetamorphosisTransforms.getClassInternalNamesToTransform()
                    .map { Target.targetClass(it) }
                    .toSet()
        }


    }
}